@@ -1,29 +1,54 @@
 # Copyright (c) 2019 Sean Vig
 
-from typing import Any, List, Tuple, Union
+from __future__ import annotations
+
+import contextlib
+from typing import Iterator, Union
 
 from pywayland.server import Display
 
 from wlroots import ffi, lib, Ptr
 from wlroots.backend import Backend
-from wlroots.wlr_types import Box, Matrix, Texture
+from wlroots.util.box import Box
+from wlroots.wlr_types import Matrix, Texture
 
-ColorType = Union[List, Tuple, ffi.CData]
+ColorType = Union[list, tuple, ffi.CData]
 
 
 class Renderer(Ptr):
-    def __init__(self, backend: Backend, display: Display) -> None:
+    def __init__(self, ptr) -> None:
         """Obtains the renderer this backend is using
 
         The renderer is automatically destroyed as the backend is destroyed.
+        """
+        self._ptr = ptr
+
+    @classmethod
+    def autocreate(cls, backend: Backend) -> Renderer:
+        """Creates a suitable renderer for a backend."""
+        ret = lib.wlr_renderer_autocreate(backend._ptr)
+        if not ret:
+            raise RuntimeError("Unable to create a renderer.")
+        return Renderer(ret)
+
+    def init_display(self, display: Display) -> None:
+        """Creates necessary shm and invokes the initialization of the implementation
 
-        :param backend:
-            The wlroots backend to get the renderer for.
         :param display:
             The Wayland display to initialize the renderer against.
         """
-        self._ptr: Any = lib.wlr_backend_get_renderer(backend._ptr)
-        lib.wlr_renderer_init_wl_display(self._ptr, display._ptr)
+        ret = lib.wlr_renderer_init_wl_display(self._ptr, display._ptr)
+        if not ret:
+            raise RuntimeError("Unable to initialize renderer for display")
+
+    @contextlib.contextmanager
+    def render(self, width: int, height: int) -> Iterator[Renderer]:
+        """Render within the generated context"""
+        self.begin(width, height)
+        try:
+            yield self
+        finally:
+            self.end()
 
     def begin(self, width: int, height: int) -> None:
         """Begin rendering with the given height and width"""
@@ -66,3 +91,31 @@ class Renderer(Ptr):
         if not isinstance(color, ffi.CData):
             color = ffi.new("float[4]", color)
         lib.wlr_render_rect(self._ptr, box._ptr, color, projection._ptr)
+
+    def scissor(self, box: Box | None) -> None:
+        """
+        Defines a scissor box. Only pixels that lie within the scissor box can be
+        modified by drawing functions. Providing a NULL `box` disables the scissor
+        box.
+        """
+        lib.wlr_renderer_scissor(self._ptr, box._ptr if box else ffi.NULL)
+
+
+class DRMFormatSet(Ptr):
+    """struct wlr_drm_format_set"""
+
+    def __init__(self, ptr) -> None:
+        self._ptr = ptr
+
+    def get(self, format: int) -> DRMFormat | None:
+        ptr = lib.wlr_drm_format_set_get(self._ptr, format)
+        if ptr == ffi.NULL:
+            return None
+        return DRMFormat(ptr)
+
+
+class DRMFormat(Ptr):
+    """struct wlr_drm_format"""
+
+    def __init__(self, ptr) -> None:
+        self._ptr = ptr
