@@ -1,15 +1,17 @@
 # Copyright Sean Vig (c) 2020
 
-from wlroots import ffi, lib
+from __future__ import annotations
+
+from wlroots import ffi, lib, Ptr
 
 
-class Timespec:
+class Timespec(Ptr):
     def __init__(self, ptr) -> None:
         """A wrapper aronud a timespec struct"""
         self._ptr = ptr
 
     @classmethod
-    def get_monotonic_time(cls) -> "Timespec":
+    def get_monotonic_time(cls) -> Timespec:
         """Get the current monotonic time"""
         timespec = ffi.new("struct timespec *")
         ret = lib.clock_gettime(lib.CLOCK_MONOTONIC, timespec)
@@ -25,7 +27,7 @@ class Timespec:
     @property
     def nsec(self) -> int:
         """Nanoseconds in designated time"""
-        return self._ptr.tv_sec
+        return self._ptr.tv_nsec
 
     @property
     def time(self) -> float:
