@@ -1,10 +1,12 @@
 # Copyright (c) Sean Vig 2019
+from __future__ import annotations
 
 import enum
 import weakref
 
-from .keyboard import Keyboard
-from wlroots import ffi, lib, Ptr
+from pywayland.server import Signal
+
+from wlroots import ffi, PtrHasData, lib
 
 _weakkeydict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()
 
@@ -25,7 +27,7 @@ class InputDeviceType(enum.IntEnum):
     SWITCH = lib.WLR_INPUT_DEVICE_SWITCH
 
 
-class InputDevice(Ptr):
+class InputDevice(PtrHasData):
     def __init__(self, ptr) -> None:
         """Create the input device from the given cdata
 
@@ -34,24 +36,32 @@ class InputDevice(Ptr):
         """
         self._ptr = ffi.cast("struct wlr_input_device *", ptr)
 
+        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
+
     @property
-    def device_type(self) -> InputDeviceType:
+    def type(self) -> InputDeviceType:
         """The device type associated with the current device"""
         return InputDeviceType(self._ptr.type)
 
     @property
-    def keyboard(self) -> Keyboard:
-        """Return the keyboard type associated with the input device
+    def vendor(self) -> int:
+        return self._ptr.vendor
 
-        The device must be a keyboard, otherwise this throws a ValueError.
+    @property
+    def product(self) -> int:
+        return self._ptr.product
+
+    @property
+    def name(self) -> str:
+        return ffi.string(self._ptr.name).decode()
+
+    def libinput_get_device_handle(self) -> ffi.CData | None:
         """
-        if self.device_type != InputDeviceType.KEYBOARD:
-            raise ValueError(
-                f"Expected keyborad input device type, got: {self.device_type}"
-            )
+        Returns the underlying libinput device if there is one.
 
-        keyboard = Keyboard(self._ptr.keyboard)
-
-        _weakkeydict[keyboard] = self._ptr
-
-        return keyboard
+        Returns a pointer to a `struct libinput_device` for use by libinput-interfacing
+        code.
+        """
+        if lib.wlr_input_device_is_libinput(self._ptr):
+            return lib.wlr_libinput_get_device_handle(self._ptr)
+        return None
