@@ -1,12 +1,15 @@
 # Copyright Sean Vig (c) 2020
 
+from __future__ import annotations
+
 import enum
 from weakref import WeakKeyDictionary
 
 from pywayland.server import Signal
 from pywayland.protocol.wayland import WlKeyboard
 
-from wlroots import ffi, lib, Ptr
+from wlroots import ffi, PtrHasData, lib, Ptr
+from wlroots.wlr_types.input_device import InputDevice
 
 _weakkeydict: WeakKeyDictionary = WeakKeyDictionary()
 
@@ -37,7 +40,7 @@ class KeyboardKeyEvent(Ptr):
         This event is emitted before the xkb state of the keyboard has been
         updated (including modifiers).
         """
-        self._ptr = ffi.cast("struct wlr_event_keyboard_key *", ptr)
+        self._ptr = ffi.cast("struct wlr_keyboard_key_event *", ptr)
 
     @property
     def time_msec(self) -> int:
@@ -60,7 +63,7 @@ class KeyboardKeyEvent(Ptr):
         return WlKeyboard.key_state(self._ptr.state)
 
 
-class Keyboard(Ptr):
+class Keyboard(PtrHasData):
     def __init__(self, ptr) -> None:
         """The Keyboard wlroots object
 
@@ -79,7 +82,17 @@ class Keyboard(Ptr):
         self.modifiers_event = Signal(ptr=ffi.addressof(self._ptr.events.modifiers))
         self.keymap_event = Signal(ptr=ffi.addressof(self._ptr.events.keymap))
         self.repeat_info_event = Signal(ptr=ffi.addressof(self._ptr.events.repeat_info))
-        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
+
+    @property
+    def base(self) -> InputDevice:
+        device_ptr = ffi.addressof(self._ptr.base)
+        _weakkeydict[device_ptr] = self._ptr
+        return InputDevice(device_ptr)
+
+    @classmethod
+    def from_input_device(cls, input_device: InputDevice) -> Keyboard:
+        ptr = lib.wlr_keyboard_from_input_device(input_device._ptr)
+        return cls(ptr)
 
     def set_keymap(self, keymap) -> None:
         """Set the keymap associated with the keyboard"""
@@ -106,7 +119,7 @@ class Keyboard(Ptr):
         return self._ptr.num_keycodes
 
     @property
-    def modifiers(self) -> "KeyboardModifiers":
+    def modifiers(self) -> KeyboardModifiers:
         """The modifiers associated with the keyboard"""
         modifiers_ptr = ffi.addressof(self._ptr.modifiers)
         _weakkeydict[modifiers_ptr] = self._ptr
@@ -115,6 +128,8 @@ class Keyboard(Ptr):
     @property
     def modifier(self) -> KeyboardModifier:
         """The enum representing the currently active modifier keys"""
+        if self._ptr == ffi.NULL:
+            raise RuntimeError("Tried to get modifier for NULL keyboard.")
         modifiers = lib.wlr_keyboard_get_modifiers(self._ptr)
         return KeyboardModifier(modifiers)
 
