@@ -4,8 +4,8 @@ from weakref import WeakKeyDictionary
 
 from pywayland.server import Display, Signal
 
-from .input_device import InputDevice
 from wlroots import ffi, lib, Ptr
+from wlroots.wlr_types.keyboard import Keyboard
 
 _weakkeydict: WeakKeyDictionary = WeakKeyDictionary()
 
@@ -16,11 +16,10 @@ class VirtualKeyboardManagerV1(Ptr):
         self._ptr = lib.wlr_virtual_keyboard_manager_v1_create(display._ptr)
 
         self.new_virtual_keyboard_event = Signal(
-            ptr=ffi.addressof(self._ptr.events.new_virtual_keyboard), data_wrapper=VirtualKeyboardV1
-        )
-        self.destroy_event = Signal(
-            ptr=ffi.addressof(self._ptr.events.destroy)
+            ptr=ffi.addressof(self._ptr.events.new_virtual_keyboard),
+            data_wrapper=VirtualKeyboardV1,
         )
+        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
 
 
 class VirtualKeyboardV1(Ptr):
@@ -28,15 +27,11 @@ class VirtualKeyboardV1(Ptr):
         """A wlr_virtual_keyboard_v1 instance."""
         self._ptr = ffi.cast("struct wlr_virtual_keyboard_v1 *", ptr)
 
-        self.destroy_event = Signal(
-            ptr=ffi.addressof(self._ptr.events.destroy), data_wrapper=VirtualKeyboardV1
-        )
-
     @property
-    def input_device(self) -> InputDevice:
-        device_ptr = ffi.addressof(self._ptr.input_device)
-        _weakkeydict[device_ptr] = self._ptr
-        return InputDevice(device_ptr)
+    def keyboard(self) -> Keyboard:
+        keyboard_ptr = ffi.addressof(self._ptr.keyboard)
+        _weakkeydict[keyboard_ptr] = self._ptr
+        return Keyboard(keyboard_ptr)
 
     @property
     def has_keymap(self) -> bool:
