@@ -1,14 +1,16 @@
 # Copyright (c) Sean Vig 2019
 
+from __future__ import annotations
+
 import enum
 import weakref
-from typing import Callable, Optional, Tuple, TypeVar
+from typing import Callable, TypeVar
 
 from pywayland.server import Display, Signal
 
-from wlroots import ffi, lib, Ptr
+from wlroots import ffi, PtrHasData, lib, Ptr, str_or_none
+from wlroots.util.box import Box
 from wlroots.util.edges import Edges
-from .box import Box
 from .output import Output
 from .surface import Surface
 
@@ -32,14 +34,21 @@ class XdgSurfaceRole(enum.IntEnum):
     POPUP = lib.WLR_XDG_SURFACE_ROLE_POPUP
 
 
-class XdgShell(Ptr):
-    def __init__(self, display: Display) -> None:
+class XdgTopLevelWMCapabilities(enum.IntFlag):
+    WINDOW_MENU = lib.WLR_XDG_TOPLEVEL_WM_CAPABILITIES_WINDOW_MENU
+    MAXIMIZE = lib.WLR_XDG_TOPLEVEL_WM_CAPABILITIES_MAXIMIZE
+    FULLSCREEN = lib.WLR_XDG_TOPLEVEL_WM_CAPABILITIES_FULLSCREEN
+    MINIMIZE = lib.WLR_XDG_TOPLEVEL_WM_CAPABILITIES_MINIMIZE
+
+
+class XdgShell(PtrHasData):
+    def __init__(self, display: Display, version: int = 5) -> None:
         """Create the shell for protocol windows
 
         :param display:
             The Wayland server display to create the shell on.
         """
-        self._ptr = lib.wlr_xdg_shell_create(display._ptr)
+        self._ptr = lib.wlr_xdg_shell_create(display._ptr, version)
 
         self.new_surface_event = Signal(
             ptr=ffi.addressof(self._ptr.events.new_surface), data_wrapper=XdgSurface
@@ -47,7 +56,7 @@ class XdgShell(Ptr):
         self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
 
 
-class XdgSurface(Ptr):
+class XdgSurface(PtrHasData):
     def __init__(self, ptr) -> None:
         """A user interface element requiring management by the compositor
 
@@ -65,10 +74,23 @@ class XdgSurface(Ptr):
         self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
         self.unmap_event = Signal(ptr=ffi.addressof(self._ptr.events.unmap))
         self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
+        self.new_popup_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.new_popup), data_wrapper=XdgPopup
+        )
+        self.configure_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.configure),
+            data_wrapper=XdgSurfaceConfigure,
+        )
+        self.ack_configure_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.ack_configure),
+            data_wrapper=XdgSurfaceConfigure,
+        )
 
     @classmethod
-    def from_surface(cls, surface: Surface) -> "XdgSurface":
+    def from_surface(cls, surface: Surface) -> XdgSurface:
         """Get the xdg surface associated with the given surface"""
+        if not surface.is_xdg_surface:
+            raise RuntimeError("Surface is not XDG surface")
         surface_ptr = lib.wlr_xdg_surface_from_wlr_surface(surface._ptr)
         return XdgSurface(surface_ptr)
 
@@ -83,7 +105,7 @@ class XdgSurface(Ptr):
         return XdgSurfaceRole(self._ptr.role)
 
     @property
-    def toplevel(self) -> "XdgTopLevel":
+    def toplevel(self) -> XdgTopLevel:
         """Return the top level xdg object
 
         This shell must be a top level role
@@ -95,10 +117,24 @@ class XdgSurface(Ptr):
 
         # the toplevel does not own the ptr data, ensure the underlying cdata
         # is kept alive
-        _weakkeydict[toplevel] = self._ptr
+        _weakkeydict[toplevel] = self
 
         return toplevel
 
+    @property
+    def popup(self) -> XdgPopup:
+        """Return the popup xdg object
+
+        This shell must be a popup role.
+        """
+        if self.role != XdgSurfaceRole.POPUP:
+            raise ValueError(f"xdg surface must be popup, got: {self.role}")
+
+        popup = XdgPopup(self._ptr.popup)
+        _weakkeydict[popup] = self
+
+        return popup
+
     def get_geometry(self) -> Box:
         """Get the surface geometry
 
@@ -112,26 +148,38 @@ class XdgSurface(Ptr):
         return Box(box_ptr.x, box_ptr.y, box_ptr.width, box_ptr.height)
 
     def set_size(self, width: int, height: int) -> int:
-        return lib.wlr_xdg_toplevel_set_size(self._ptr, width, height)
+        return lib.wlr_xdg_toplevel_set_size(self._ptr.toplevel, width, height)
 
     def set_activated(self, activated: bool) -> int:
         if self.role != XdgSurfaceRole.TOPLEVEL:
             raise ValueError(f"xdg surface must be top-level, got: {self.role}")
 
-        return lib.wlr_xdg_toplevel_set_activated(self._ptr, activated)
+        return lib.wlr_xdg_toplevel_set_activated(self._ptr.toplevel, activated)
 
-    def set_tiled(self, tiled: int) -> int:
-        return lib.wlr_xdg_toplevel_set_tiled(self._ptr, tiled)
+    def set_maximized(self, maximized: bool) -> int:
+        return lib.wlr_xdg_toplevel_set_maximized(self._ptr.toplevel, maximized)
 
-    def set_fullscreen(self, fullscreened: bool) -> int:
-        return lib.wlr_xdg_toplevel_set_fullscreen(self._ptr, fullscreened)
+    def set_fullscreen(self, fullscreen: bool) -> int:
+        return lib.wlr_xdg_toplevel_set_fullscreen(self._ptr.toplevel, fullscreen)
+
+    def set_resizing(self, resizing: bool) -> int:
+        return lib.wlr_xdg_toplevel_set_resizing(self._ptr.toplevel, resizing)
+
+    def set_tiled(self, tiled_edges: int) -> int:
+        return lib.wlr_xdg_toplevel_set_tiled(self._ptr.toplevel, tiled_edges)
+
+    def set_bounds(self, width: int, height: int) -> int:
+        return lib.wlr_xdg_toplevel_set_bounds(self._ptr.toplevel, width, height)
+
+    def set_wm_capabilities(self, caps: XdgTopLevelWMCapabilities) -> int:
+        return lib.wlr_xdg_toplevel_set_wm_capabilities(self._ptr.toplevel, caps)
 
     def send_close(self) -> int:
-        return lib.wlr_xdg_toplevel_send_close(self._ptr)
+        return lib.wlr_xdg_toplevel_send_close(self._ptr.toplevel)
 
     def surface_at(
         self, surface_x: float, surface_y: float
-    ) -> Tuple[Optional[Surface], float, float]:
+    ) -> tuple[Surface | None, float, float]:
         """Find a surface within this xdg-surface tree at the given surface-local coordinates
 
         Returns the surface and coordinates in the leaf surface coordinate
@@ -148,7 +196,7 @@ class XdgSurface(Ptr):
         return Surface(surface_ptr), sub_x_data[0], sub_y_data[0]
 
     def for_each_surface(
-        self, iterator: SurfaceCallback[T], data: Optional[T] = None
+        self, iterator: SurfaceCallback[T], data: T | None = None
     ) -> None:
         """Call iterator on each surface and popup in the xdg-surface tree
 
@@ -167,6 +215,27 @@ class XdgSurface(Ptr):
             self._ptr, lib.surface_iterator_callback, handle
         )
 
+    def schedule_configure(self) -> int:
+        """
+        Schedule a surface configuration. This should only be called by protocols
+        extending the shell.
+        """
+        return lib.wlr_xdg_surface_schedule_configure(self._ptr)
+
+
+class XdgSurfaceConfigure(Ptr):
+    def __init__(self, ptr) -> None:
+        self._ptr = ffi.cast("struct wlr_xdg_surface_configure *", ptr)
+
+    @property
+    def surface(self) -> XdgSurface:
+        # TODO: keep weakref
+        return XdgSurface(self._ptr.surface)
+
+    @property
+    def serial(self) -> int:
+        return self._ptr.serial
+
 
 class XdgTopLevel(Ptr):
     def __init__(self, ptr) -> None:
@@ -182,7 +251,6 @@ class XdgTopLevel(Ptr):
         )
         self.request_fullscreen_event = Signal(
             ptr=ffi.addressof(self._ptr.events.request_fullscreen),
-            data_wrapper=XdgTopLevelSetFullscreenEvent,
         )
         self.request_minimize_event = Signal(
             ptr=ffi.addressof(self._ptr.events.request_minimize)
@@ -204,14 +272,27 @@ class XdgTopLevel(Ptr):
         self.set_app_id_event = Signal(ptr=ffi.addressof(self._ptr.events.set_app_id))
 
     @property
-    def title(self) -> str:
-        """The title of the toplevel object"""
-        return ffi.string(self._ptr.title).decode()
+    def parent(self) -> XdgTopLevel | None:
+        """The parent of this toplevel"""
+        parent_ptr = self._ptr.parent
+        if parent_ptr is None:
+            return None
+        return XdgTopLevel(parent_ptr)
 
     @property
-    def app_id(self) -> str:
+    def title(self) -> str | None:
+        """The title of the toplevel object"""
+        return str_or_none(self._ptr.title)
+
+    @property
+    def app_id(self) -> str | None:
         """The app id of the toplevel object"""
-        return ffi.string(self._ptr.app_id).decode()
+        return str_or_none(self._ptr.app_id)
+
+    @property
+    def requested(self) -> XdgTopLevelRequested:
+        """Requested initial state"""
+        return XdgTopLevelRequested(self._ptr.requested)
 
 
 class XdgTopLevelMoveEvent(Ptr):
@@ -219,9 +300,9 @@ class XdgTopLevelMoveEvent(Ptr):
         self._ptr = ffi.cast("struct wlr_xdg_toplevel_move_event *", ptr)
 
     @property
-    def surface(self) -> Surface:
+    def toplevel(self) -> XdgTopLevel:
         # TODO: keep weakref
-        return Surface(self._ptr.surface)
+        return XdgTopLevel(self._ptr.toplevel)
 
     # TODO: seat client
 
@@ -235,9 +316,9 @@ class XdgTopLevelResizeEvent(Ptr):
         self._ptr = ffi.cast("struct wlr_xdg_toplevel_resize_event *", ptr)
 
     @property
-    def surface(self) -> Surface:
+    def toplevel(self) -> XdgTopLevel:
         # TODO: keep weakref
-        return Surface(self._ptr.surface)
+        return XdgTopLevel(self._ptr.toplevel)
 
     # TODO: seat client
 
@@ -250,32 +331,14 @@ class XdgTopLevelResizeEvent(Ptr):
         return self._ptr.edges
 
 
-class XdgTopLevelSetFullscreenEvent(Ptr):
-    def __init__(self, ptr) -> None:
-        self._ptr = ffi.cast("struct wlr_xdg_toplevel_set_fullscreen_event *", ptr)
-
-    @property
-    def surface(self) -> Surface:
-        # TODO: keep weakref
-        return Surface(self._ptr.surface)
-
-    @property
-    def fullscreen(self) -> bool:
-        return self._ptr.fullscreen
-
-    @property
-    def output(self) -> Output:
-        return Output(self._ptr.output)
-
-
 class XdgTopLevelShowWindowMenuEvent(Ptr):
     def __init__(self, ptr) -> None:
         self._ptr = ffi.cast("struct wlr_xdg_toplevel_show_window_menu_event *", ptr)
 
     @property
-    def surface(self) -> Surface:
+    def toplevel(self) -> XdgTopLevel:
         # TODO: keep weakref
-        return Surface(self._ptr.surface)
+        return XdgTopLevel(self._ptr.toplevel)
 
     # TODO: seat client
 
@@ -290,3 +353,93 @@ class XdgTopLevelShowWindowMenuEvent(Ptr):
     @property
     def y(self) -> int:
         return self._ptr.y
+
+
+class XdgPopup(Ptr):
+    def __init__(self, ptr) -> None:
+        """A wlr_xdg_popup
+
+        :param ptr:
+            The wlr_xdg_popup cdata pointer
+        """
+        self._ptr = ffi.cast("struct wlr_xdg_popup *", ptr)
+
+        self.reposition_event = Signal(ptr=ffi.addressof(self._ptr.events.reposition))
+
+    @property
+    def base(self) -> XdgSurface:
+        """The xdg surface associated with the popup"""
+        return XdgSurface(self._ptr.base)
+
+    @property
+    def parent(self) -> Surface:
+        """Parent Surface."""
+        parent_ptr = self._ptr.parent
+        _weakkeydict[parent_ptr] = self
+        return Surface(parent_ptr)
+
+    @property
+    def current(self) -> XdgPopupState:
+        return XdgPopupState(self._ptr.current)
+
+    @property
+    def pending(self) -> XdgPopupState:
+        return XdgPopupState(self._ptr.pending)
+
+    def unconstrain_from_box(self, box: Box) -> None:
+        """
+        Set the geometry of this popup to unconstrain it according to its xdg-positioner
+        rules. The box should be in the popup's root toplevel parent surface coordinate
+        system.
+        """
+        lib.wlr_xdg_popup_unconstrain_from_box(self._ptr, box._ptr)
+
+    def destroy(self) -> None:
+        """Request that this popup closes."""
+        lib.wlr_xdg_popup_destroy(self._ptr)
+
+
+class XdgPopupState(Ptr):
+    def __init__(self, ptr) -> None:
+        """A struct wlr_xdg_popup_state
+
+        :param ptr:
+            The wlr_xdg_popup_state cdata pointer
+        """
+        self._ptr = ffi.cast("struct wlr_xdg_popup_state *", ptr)
+
+    @property
+    def geometry(self) -> Box:
+        """
+        Position of the popup relative to the upper left corner of the window geometry
+        of the parent surface.
+        """
+        return Box(self._ptr.geometry)
+
+    @property
+    def reactive(self) -> bool:
+        return self._ptr.reactive
+
+
+class XdgTopLevelRequested(Ptr):
+    def __init__(self, ptr) -> None:
+        self._ptr = ptr
+
+    @property
+    def maximized(self) -> bool:
+        return self._ptr.maximized
+
+    @property
+    def minimized(self) -> bool:
+        return self._ptr.minimized
+
+    @property
+    def fullscreen(self) -> bool:
+        return self._ptr.fullscreen
+
+    @property
+    def fullscreen_output(self) -> Output | None:
+        output_ptr = self._ptr.fullscreen_output
+        if output_ptr == ffi.NULL:
+            return None
+        return Output(output_ptr)
